@@ -32,7 +32,7 @@
 extern void *C_LoadModule(const char *name, CK_FUNCTION_LIST_PTR_PTR);
 extern CK_RV C_UnloadModule(void *module);
 
-#if OPENSSL_VERSION_NUMBER < 0x10100004L
+#if OPENSSL_VERSION_NUMBER < 0x10100004L || defined(LIBRESSL_VERSION_NUMBER)
 typedef int PKCS11_RWLOCK;
 #else
 typedef CRYPTO_RWLOCK *PKCS11_RWLOCK;
@@ -47,6 +47,8 @@ typedef struct pkcs11_ctx_private {
 	CK_FUNCTION_LIST_PTR method;
 	void *handle;
 	char *init_args;
+	UI_METHOD *ui_method; /* UI_METHOD for CKU_CONTEXT_SPECIFIC PINs */
+	void *ui_user_data;
 	unsigned int forkid;
 	PKCS11_RWLOCK rwlock;
 } PKCS11_CTX_private;
@@ -63,9 +65,6 @@ typedef struct pkcs11_slot_private {
 	/* options used in last PKCS11_login */
 	char *prev_pin;
 	int prev_so;
-
-	/* per-slot lock */
-	PKCS11_RWLOCK rwlock;
 } PKCS11_SLOT_private;
 #define PRIVSLOT(slot)		((PKCS11_SLOT_private *) ((slot)->_private))
 #define SLOT2CTX(slot)		(PRIVSLOT(slot)->parent)
@@ -88,6 +87,7 @@ typedef struct pkcs11_token_private {
 typedef struct pkcs11_key_ops {
 	int type; /* EVP_PKEY_xxx */
 	EVP_PKEY *(*get_evp_key) (PKCS11_KEY *);
+	void (*update_ex_data) (PKCS11_KEY *);
 } PKCS11_KEY_ops;
 
 typedef struct pkcs11_key_private {
@@ -115,34 +115,26 @@ typedef struct pkcs11_cert_private {
 #define CERT2TOKEN(cert)	(PRIVCERT(cert)->parent)
 #define CERT2CTX(cert)		TOKEN2CTX(CERT2TOKEN(cert))
 
-/*
- * Mapping Cryptoki error codes to those used internally
- * by this code.
- * Right now, we just map them directly, and make sure
- * that the few genuine messages we use don't clash with
- * PKCS#11
- */
-#define pkcs11_map_err(rv)	(rv)
-
 /*
  * Internal functions
  */
 #define CRYPTOKI_checkerr(f, rv) \
 	do { \
 		if (rv) { \
-			PKCS11err(f, pkcs11_map_err(rv)); \
+			CKRerr(f, rv); \
 			return -1; \
 		} \
 		ERR_clear_error(); \
 	} while (0)
 #define CRYPTOKI_call(ctx, func_and_args) \
 	PRIVCTX(ctx)->method->func_and_args
+extern int ERR_load_CKR_strings(void);
 
 /* Memory allocation */
 #define PKCS11_DUP(s) \
 	pkcs11_strdup((char *) s, sizeof(s))
 
-#if OPENSSL_VERSION_NUMBER < 0x10100004L
+#if OPENSSL_VERSION_NUMBER < 0x10100004L || defined(LIBRESSL_VERSION_NUMBER)
 /* Emulate the OpenSSL 1.1 locking API for older OpenSSL versions */
 int CRYPTO_THREAD_lock_new();
 void CRYPTO_THREAD_lock_free(int);
@@ -274,6 +266,10 @@ extern PKCS11_KEY *pkcs11_find_key_from_key(PKCS11_KEY *key);
 extern int pkcs11_enumerate_certs(PKCS11_TOKEN *token,
 	PKCS11_CERT **certs, unsigned int *ncerts);
 
+/* Set UI method to allow retrieving CKU_CONTEXT_SPECIFIC PINs interactively */
+extern int pkcs11_set_ui_method(PKCS11_CTX *ctx,
+	UI_METHOD *ui_method, void *ui_user_data);
+
 /* Initialize a token */
 extern int pkcs11_init_token(PKCS11_TOKEN * token, const char *pin,
 	const char *label);
